/*
 * dbcache.c, PFC d'en Guillem Cantallops Ramis a la UIB
 *
 * Copyright (C) 2000-2001, GCR.
 *
 * Distributed under the terms of the GNU GPL
 *
 */

#define __USE_GNU

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <mysql/mysql.h>
#include <pthread.h>
#include <string.h>
#include "dbcache.h"




/****** Variables globals ******/

 
static char				query[MAX_QUERY_LEN];				/* buffer per posar les peticions a la BD */
static MYSQL			mysql;								/* descriptor de connexi amb la BD */
static MYSQL_RES		*result = NULL;						/* punter a un resultat de la BD */
static MYSQL_ROW		row;								/* estructura per una fila tornada per la BD */
static pthread_mutex_t	mdb = PTHREAD_MUTEX_INITIALIZER;	/* mutex per accs a la BD */
															/* es OBLIGATORI entre mysql_query() i mysql_store_result() */
															/* pero tambe l'utilitzam per protegir aquestes variables */
															/* EXCEPTE "row" que s usada a un sol thread. ALERTA!!! */

static NodeHost			nh;									/* estructura i taula de hash per HOST, amb mutex */
static NodeHost			*hh[HASH_SIZE];
static pthread_mutex_t	mh = PTHREAD_MUTEX_INITIALIZER;
static u_int			rhh = 0, rmh = 0;					/* per estadistiques (read hit host, read miss host) */
static float			rh_ea = 0.0;
static NodeHost			*clhandh = NULL;					/* la maneta del CLOCK */
static u_int			htlenh = 0;							/* numero d'elements a la cache de la taula */

static NodeObert		no;									/* estructura i taula de hash per OBERT, amb mutex */
static NodeObert		*ho[HASH_SIZE];
static pthread_mutex_t	mo = PTHREAD_MUTEX_INITIALIZER;
static u_int			rho = 0, rmo = 0;					/* per estadistiques (read hit obert, read miss obert) */
static float			ro_ea = 0.0;
static NodeObert		*clhando = NULL;					/* la maneta del CLOCK */
static u_int			htleno = 0;							/* numero d'elements a la cache de la taula */

static NodeServei		ns;									/* estructura i taula de hash SERVEI, sense mutex */
static NodeServei		*hs[HASH_SIZE];						/* (no t mutex perqu s "read only", no t stats) */

extern int				flag_acabar;						/* flag global per controlar l'acabament, amb mutex */
extern pthread_mutex_t	mf;

extern int				errno;




/****** Funcions de depuraci - Thread "principal" ******/


#ifdef DEBUG

/* NOTA: Algorismes basats en void gravador(void), que est ben comentat al final d'aquest fitxer */

inline void dump_host_cache (void)
{
	register u_long			hash;
	register NodeHost		*tmp;
 
	printf("dump_host_cache:\n");
	for(hash=0; hash<HASH_SIZE; hash++) {
		if(!(tmp=hh[hash])) continue;
		printf("%lu:\t", hash);
		while(tmp) {
			printf("(%lu,%lu,%lu) ", tmp->ip, tmp->mach, tmp->macl);
			tmp=tmp->next;
		}
		printf("\n");
	}
}


inline void dump_obert_cache (void)
{
	register u_long			hash;
	register NodeObert		*tmp;
 
	printf("dump_obert_cache:\n");
	for(hash=0; hash<HASH_SIZE; hash++) {
		if(!(tmp=ho[hash])) continue;
		printf("%lu:\t", hash);
		while(tmp) {
			printf("(%lu,%lu,%lu,%u,%u) ", tmp->ip, tmp->mach, tmp->macl, tmp->port, tmp->protocol);
			tmp=tmp->next;
		}
		printf("\n");
	}
}


inline void dump_servei_cache (void)
{
	register u_long			hash;
	register NodeServei		*tmp;

	printf("dump_servei_cache:\n");
	for(hash=0; hash<HASH_SIZE; hash++) {
		if(!(tmp=hs[hash])) continue;
		printf("%lu:\t", hash);
		while(tmp) {
			printf("(%u,%u) ", tmp->port, tmp->protocol);
			tmp=tmp->next;
		}
		printf("\n");
	}
}

#endif




/****** Funcions d'inicialitzaci i tancament - Thread "principal" ******/


/* Inicialitza variables globals i assigna recursos */
inline void init_dbcache (void)
{
	/* Netejam les taules de hash i les estructures de suport (els mutex ja estan inicialitzats) */
	memset(&nh, 0, sizeof(nh));		/* init_node_host() */
	memset(hh, 0, sizeof(hh));		/* init_hash_host() */
	memset(&no, 0, sizeof(no));		/* init_node_obert() */
	memset(ho, 0, sizeof(ho));		/* init_hash_obert() */
	memset(&ns, 0, sizeof(ns));		/* init_node_servei() */
	memset(hs, 0, sizeof(hs));		/* init_hash_servei() */

	/* MySQL: inicialitzam l'estructura i ens connectam a la BD */
#ifdef DEBUG
	printf("mysql_init: ");
#endif
	if(!mysql_init(&mysql))
		MYSQL_ERR(&mysql, "mysql_init");
#ifdef DEBUG
	printf("ok\n");
#endif
#ifdef DEBUG
	printf("mysql_real_connect: ");
#endif
	if(!mysql_real_connect(&mysql, HOST, USER, PASSWD, DB, PORT, UNIX_SOCKET, CLIENT_FLAG))
		MYSQL_ERR(&mysql, "mysql_real_connect")
#ifdef DEBUG
	printf("ok\n");
#endif
}


/* Allibera recursos */
inline void close_dbcache (void)
{
#ifdef DEBUG
	dump_host_cache();
	dump_obert_cache();
	dump_servei_cache();
#endif
	/* MySQL: alliberam recursos i desconnectam */
#ifdef DEBUG
	printf("mysql_close: ");
#endif
	mysql_close(&mysql);
#ifdef DEBUG
	printf("ok\n");
#endif 
}




/****** Funcions internes de la llibreria - Base de Dades (_NOMES_ gestionen el mutex de la BD) ******/


inline void db_insert_host (NodeHost *tmph)
{
	pthread_mutex_lock(&mdb);
	sprintf(
		query,
		"insert into host (ip, mach, macl, omega, alfa) \
		values (%lu, %lu, %lu, from_unixtime(%lu), from_unixtime(%lu))",
		tmph->ip, tmph->mach, tmph->macl, tmph->omega, tmph->alfa
	);
	if(mysql_query(&mysql, query))				/* Fem la petici */
		MYSQL_ERR(&mysql, "mysql_query")
	pthread_mutex_unlock(&mdb);
}


inline void db_update_omega_host (NodeHost *tmph)
{
	pthread_mutex_lock(&mdb);
	sprintf(
		query,
		"update host set omega=from_unixtime(%lu) \
		where ip=%lu and mach=%lu and macl=%lu",
		tmph->omega, tmph->ip, tmph->mach, tmph->macl
	);
	if(mysql_query(&mysql, query))				/* Fem la petici */
		MYSQL_ERR(&mysql, "mysql_query")
	pthread_mutex_unlock(&mdb);
}


inline void db_insert_obert (NodeObert *tmpo)
{
	pthread_mutex_lock(&mdb);
	sprintf(
		query,
		"insert into obert (ip, mach, macl, port, protocol, omega, alfa) \
		values (%lu, %lu, %lu, %u, %u, from_unixtime(%lu), from_unixtime(%lu))",
		tmpo->ip, tmpo->mach, tmpo->macl, tmpo->port, tmpo->protocol, tmpo->omega, tmpo->alfa
	);
	if(mysql_query(&mysql, query))				/* Fem la petici */
		MYSQL_ERR(&mysql, "mysql_query")
	pthread_mutex_unlock(&mdb);
}


inline void db_update_omega_obert (NodeObert *tmpo)
{
	pthread_mutex_lock(&mdb);
	sprintf(
		query,
		"update obert set omega=from_unixtime(%lu) \
		where ip=%lu and mach=%lu and macl=%lu and port=%u and protocol=%u",
		tmpo->omega, tmpo->ip, tmpo->mach, tmpo->macl, tmpo->port, tmpo->protocol
	);
	if(mysql_query(&mysql, query))				/* Fem la petici */
		MYSQL_ERR(&mysql, "mysql_query")
	pthread_mutex_unlock(&mdb);
}


/* ALERTA que agafa dades de la global "nh" (_no_ gestiona
   el seu mutex) i torna el resultat a dins la global "row" */
inline void db_lookup_host (void)
{
	pthread_mutex_lock(&mdb);
	sprintf(
		query,
		"select unix_timestamp(omega), unix_timestamp(alfa) from host where ip=%lu and mach=%lu and macl=%lu",
		nh.ip, nh.mach, nh.macl
	);
	if(mysql_query(&mysql, query))				/* Fem la petici */
		MYSQL_ERR(&mysql, "mysql_query")
	if(!(result=mysql_store_result(&mysql)))	/* Llegim el resultat */
		MYSQL_ERR(&mysql, "mysql_store_result")
	row=mysql_fetch_row(result);				/* Agafam l'unica fila que hauria de tornar com a molt */
	mysql_free_result(result);					/* Ja podem alliberar els recursos ocupats pel resultat */
	pthread_mutex_unlock(&mdb);
}


/* ALERTA que agafa dades de la global "no" (_no_ gestiona
   el seu mutex) i torna el resultat a dins la global "row" */
inline void db_lookup_obert (void)
{
	pthread_mutex_lock(&mdb);
	sprintf(
		query,
		"select unix_timestamp(omega), unix_timestamp(alfa) from obert \
			where ip=%lu and mach=%lu and macl=%lu and port=%u and protocol=%u",
		no.ip, no.mach, no.macl, no.port, no.protocol
	);
	if(mysql_query(&mysql, query))				/* Fem la petici */
		MYSQL_ERR(&mysql, "mysql_query")
	if(!(result=mysql_store_result(&mysql)))	/* Llegim el resultat */
		MYSQL_ERR(&mysql, "mysql_store_result")
	row=mysql_fetch_row(result);				/* Agafam l'unica fila que hauria de tornar com a molt */
	mysql_free_result(result);					/* Ja podem alliberar els recursos ocupats pel resultat */
	pthread_mutex_unlock(&mdb);
}


/* ALERTA que agafa dades de la global "ns" (_no_ gestiona
   el seu mutex) i torna el resultat a dins la global "row" */
inline void db_lookup_servei (void)
{
	pthread_mutex_lock(&mdb);
	sprintf(
		query,
		"select 1 from servei where port=%u and protocol=%u",
		ns.port, ns.protocol
	);
	if(mysql_query(&mysql, query))				/* Fem la petici */
		MYSQL_ERR(&mysql, "mysql_query")
	if(!(result=mysql_store_result(&mysql)))	/* Llegim el resultat */
		MYSQL_ERR(&mysql, "mysql_store_result")
	row=mysql_fetch_row(result);				/* Agafam l'unica fila que hauria de tornar com a molt */
	mysql_free_result(result);					/* Ja podem alliberar els recursos ocupats pel resultat */
	pthread_mutex_unlock(&mdb);
}




/****** Funcions internes de la llibreria - Hashing (aqui no tocam els mutex, qui les crida ho ha de fer!) ******/


/* Aquesta funci espera tenir dades d'un host a dins nh.
   Intenta trobar el host a hh, i si el troba torna el punter corresponent.
   En cas contrari, torna NULL.
*/
inline NodeHost *get_host (void)
{
	register NodeHost	*tmp=hh[HH(nh.ip, nh.mach, nh.macl)];

	while (tmp) {
		if (!NHCMP(tmp, &nh)) {
			return tmp;		/* L'hem trobat, tornam un punter que diu a on */
		}
		tmp=tmp->next;		/* Encara no l'hem trobat, seguim cercant */
	}
	return tmp;				/* i.e. "return NULL;", definitivament no l'hem trobat ;-) */
}


/* Aquesta funci espera tenir dades d'un host a dins nh.
   Afegeix un nou element a hh amb les dades de nh i torna el punter.
   Si hi ha algun problema, torna NULL.
*/
inline NodeHost *add_host (void)
{
	register u_long		hash  = HH(nh.ip, nh.mach, nh.macl);
	register NodeHost	*new  = NULL;
	register u_long		hash2 = 0;
	register NodeHost	*tmp  = NULL;

	if (htlenh<MAX_HT_LEN) {						/* Si no passam la longitud maxima, afegim nou element */
		new=(NodeHost *)malloc(sizeof(NodeHost));
		memcpy(new, &nh, sizeof(NodeHost));
		new->next=hh[hash];
		hh[hash]=new;
		if (!htlenh) {								/* Primer element */
			clhandh=new;
			clhandh->clnext=clhandh;
		}
		else {										/* Un dels altres */
			new->clnext=clhandh->clnext;
			clhandh->clnext=new;
		}
		htlenh++;
		return new;
	}

	while (clhandh->used) {							/* En cas contrari seleccionam victima amb el CLOCK */
		clhandh->used=0;
		clhandh=clhandh->clnext;
	}
	new=clhandh;
	clhandh=clhandh->clnext;
	if (new->ghost) {								/* Guardam (si hi ha alguna cosa que guardar) */
		db_insert_host(new);
	}
	else if (new->dirty) {
		db_update_omega_host(new);
	}												/* I fem el divertit proces de substitucio */
	hash2=HH(new->ip, new->mach, new->macl);
	tmp=hh[hash2];
	if (tmp==new) {
		hh[hash2]=new->next;
	}
	else {
		while (tmp->next) {
			if (!NHCMP(tmp->next, new)) {
				tmp->next=new->next;
				break;
			}
			tmp=tmp->next;
		}
	}
	nh.next=hh[hash];
	hh[hash]=new;
	nh.clnext=new->clnext;
	memcpy(new, &nh, sizeof(NodeHost));
	return new;
}


/* Aquesta funci espera tenir dades d'un obert a dins no.
   Intenta trobar l'obert a ho, i si el troba torna el punter corresponent.
   En cas contrari, torna NULL.
*/
inline NodeObert *get_obert (void)
{
	register NodeObert	*tmp=ho[HO(no.ip, no.mach, no.macl, no.port, no.protocol)];

	while (tmp) {
		if (!NOCMP(tmp, &no)) {
			return tmp;		/* L'hem trobat, tornam un punter que diu a on */
		}
		tmp=tmp->next;		/* Encara no l'hem trobat, seguim cercant */
	}
	return tmp;				/* i.e. "return NULL;", definitivament no l'hem trobat ;-) */
}


/* Aquesta funci espera tenir dades d'un obert a dins no.
   Afegeix un nou element a ho amb les dades de no i torna el punter.
   Si hi ha algun problema, torna NULL.
*/
inline NodeObert *add_obert (void)
{
	register u_long		hash  = HO(no.ip, no.mach, no.macl, no.port, no.protocol);
	register NodeObert	*new  = NULL;
	register u_long		hash2 = 0;
	register NodeObert	*tmp  = NULL;

	if (htleno<MAX_HT_LEN) {						/* Si no passam la longitud maxima, afegim nou element */
		new=(NodeObert *)malloc(sizeof(NodeObert));
		memcpy(new, &no, sizeof(NodeObert));
		new->next=ho[hash];
		ho[hash]=new;
		if (!htleno) {								/* Primer element */
			clhando=new;
			clhando->clnext=clhando;
		}
		else {										/* Un dels altres */
			new->clnext=clhando->clnext;
			clhando->clnext=new;
		}
		htleno++;
		return new;
	}

	while (clhando->used) {							/* En cas contrari seleccionam victima amb el CLOCK */
		clhando->used=0;
		clhando=clhando->clnext;
	}
	new=clhando;
	clhando=clhando->clnext;
	if (new->ghost) {								/* Guardam (si hi ha alguna cosa que guardar) */
		db_insert_obert(new);
	}
	else if (new->dirty) {
		db_update_omega_obert(new);
	}												/* I fem el divertit proces de substitucio */
	hash2=HO(new->ip, new->mach, new->macl, new->port, new->protocol);
	tmp=ho[hash2];
	if (tmp==new) {
		ho[hash2]=new->next;
	}
	else {
		while (tmp->next) {
			if (!NOCMP(tmp->next, new)) {
				tmp->next=new->next;
				break;
			}
			tmp=tmp->next;
		}
	}
	no.next=ho[hash];
	ho[hash]=new;
	no.clnext=new->clnext;
	memcpy(new, &no, sizeof(NodeObert));
	return new;
}


/* Aquesta funci espera tenir dades d'un servei a dins ns.
   Intenta trobar el servei a hs, i si el troba torna el punter corresponent.
   En cas contrari, torna NULL.
*/
inline NodeServei *get_servei (void)
{
	register NodeServei	*tmp=hs[HS(ns.port, ns.protocol)];

	while (tmp) {
		if (!NSCMP(tmp, &ns)) {
			return tmp;		/* L'hem trobat, tornam un punter que diu a on */
		}
		tmp=tmp->next;		/* Encara no l'hem trobat, seguim cercant */
	}
	return tmp;				/* i.e. "return NULL;", definitivament no l'hem trobat ;-) */
}


/* Aquesta funci espera tenir dades d'un obert a dins no.
   Afegeix un nou element a ho amb les dades de no i torna el punter.
   Si hi ha algun problema, torna NULL.
*/
inline NodeServei *add_servei (void)
{
	register u_long		hash=HS(ns.port, ns.protocol);
	register NodeServei	*new;

	new=(NodeServei *)malloc(sizeof(NodeServei));
	memcpy(new, &ns, sizeof(NodeServei));
	new->next=hs[hash];
	hs[hash]=new;
	return new;				/* Valor nou, l'afegim i tornam un punter que diu a on es */
}




/****** Funcions que proporciona la llibreria - Accs a MySQL amb cach - Thread "capturador" ******/


/* Afegeix una nova entrada de host */
inline void insertar_host (u_int ip, u_int mach, u_int macl)
{
	pthread_mutex_lock(&mh);
	nh.ip=ip;				/* Omplim les dades propies de la taula */
	nh.mach=mach;
	nh.macl=macl;
	nh.omega=time(NULL);
	nh.alfa=time(NULL);
	nh.ghost=1;				/* Entrada fantasma: es a la cache pero no a la BD, requereix INSERT */
	nh.dirty=0;				/* Entrada neta: no ha estat modificada, no requereix UPDATE */
	nh.used=0;
	add_host();				/* Posam totes aquestes dades a la taula de hash corresponent */
	pthread_mutex_unlock(&mh);
}


/* Actualitza el timestamp del darrer avistament d'un parell IP/MAC */
/* NOTA: Si no existeix... pitipif! Comprovar-ho abans!!! */
inline void actualitzar_omega_host (u_int ip, u_int mach, u_int macl)
{
	register NodeHost	*tmp;

	pthread_mutex_lock(&mh);
	nh.ip=ip;				/* Especificam les claus de la taula, per cercar */
	nh.mach=mach;
	nh.macl=macl;
	tmp=get_host();			/* Obtenim la situacio de l'entrada que t aquestes claus */
	tmp->omega=time(NULL);	/* Actualitzam el timestamp */
	tmp->dirty=1;			/* Entrada bruta: ha estat modificada, requereix UPDATE */
	tmp->used=1;
	pthread_mutex_unlock(&mh);
}


/* Afegeix una nova entrada de servei obert */
inline void insertar_obert (u_int ip, u_int mach, u_int macl, u_short port, u_char protocol)
{
	pthread_mutex_lock(&mo);
	no.ip=ip;				/* Omplim les dades propies de la taula */
	no.mach=mach;
	no.macl=macl;
	no.port=port;
	no.protocol=protocol;
	no.omega=time(NULL);
	no.alfa=time(NULL);
	no.ghost=1;				/* Entrada fantasma: es a la cache pero no a la BD, requereix INSERT */
	no.dirty=0;				/* Entrada neta: no ha estat modificada, no requereix UPDATE */
	no.used=0;
	add_obert();			/* Posam totes aquestes dades a la taula de hash corresponent */
	pthread_mutex_unlock(&mo);
}


/* Actualitza el timestamp del darrer avistament d'un servei obert */
/* NOTA: Si no existeix... pitipif! Comprovar-ho abans!!! */
inline void actualitzar_omega_obert (u_int ip, u_int mach, u_int macl, u_short port, u_char protocol)
{
	register NodeObert	*tmp;

	pthread_mutex_lock(&mo);
	no.ip=ip;				/* Especificam les claus de la taula, per cercar */
	no.mach=mach;
	no.macl=macl;
	no.port=port;
	no.protocol=protocol;
	tmp=get_obert();		/* Obtenim la situacio de l'entrada que t aquestes claus */
	tmp->omega=time(NULL);	/* Actualitzam el timestamp */
	tmp->dirty=1;			/* Entrada bruta: ha estat modificada, requereix UPDATE */
	tmp->used=1;
	pthread_mutex_unlock(&mo);
}


/* Ens diu si ja tenim fitxat el host en questi */
inline int hi_es_el_host (u_int ip, u_int mach, u_int macl)
{
	register NodeHost	*tmp;

	pthread_mutex_lock(&mh);
	nh.ip=ip;									/* Especificam les claus de la taula */
	nh.mach=mach;
	nh.macl=macl;
	if ((tmp=get_host())) {						/* Cercam l'entrada corresponent a la cach */
		tmp->used=1;
		rhh++;
		pthread_mutex_unlock(&mh);
		return 1;								/* Si hi es, avisam i hem acabat */
	}											/* Si no, hem d'anar a cercar-lo a la BD */
	db_lookup_host();
	if(!row) {									/* Miram si la BD ens ha tornat alguna cosa */
		pthread_mutex_unlock(&mh);
		return 0;								/* Si no, avisam i hem acabat */
	}											/* Per si l'hem trobat */
	nh.omega=strtoul(row[0], NULL, 10);			/* Completam les dades que falten a l'estructura */
	nh.alfa=strtoul(row[1], NULL, 10);
	nh.ghost=0;
	nh.dirty=0;
	nh.used=0;
	add_host();									/* Afegim la nova entrada a la cach */
	rmh++;
	pthread_mutex_unlock(&mh);
	return 1;									/* I avisam de que hi es */
}


/* Ens diu si tenim el servei a la taula corresponent */
inline int hi_es_el_servei (u_short port, u_char protocol)
{
	ns.port=port;								/* Especificam les claus de la taula */
	ns.protocol=protocol;
	if (get_servei()) {							/* Cercam l'entrada corresponent a la cach */
		return 1;								/* Si hi es, avisam i hem acabat */
	}											/* Si no, hem d'anar a cercar-lo a la BD */
	db_lookup_servei();
	if(!row) {									/* Miram si la BD ens ha tornat alguna cosa */
		return 0;								/* Si no, avisam i hem acabat */
	}											/* Per si l'hem trobat */
	add_servei();								/* Afegim la nova entrada a la cach */
	return 1;									/* I avisam de que hi es */
}


/* Ens diu si ja tenim fitxat el servei obert en questi */
inline int esta_obert_i_ho_sabem (u_int ip, u_int mach, u_int macl, u_short port, u_char protocol)
{
	register NodeObert	*tmp;

	pthread_mutex_lock(&mo);
	no.ip=ip;									/* Especificam les claus de la taula, per cercar */
	no.mach=mach;
	no.macl=macl;
	no.port=port;
	no.protocol=protocol;
	if ((tmp=get_obert())) {							/* Cercam l'entrada corresponent a la cach */
		tmp->used=1;
		rho++;
		pthread_mutex_unlock(&mo);
		return 1;								/* Si hi es, avisam i hem acabat */
	}											/* Si no, hem d'anar a cercar-lo a la BD */
	db_lookup_obert();
	if(!row) {									/* Miram si la BD ens ha tornat alguna cosa */
		pthread_mutex_unlock(&mo);
		return 0;								/* Si no, avisam i hem acabat */
	}											/* Per si l'hem trobat */
	no.omega=strtoul(row[0], NULL, 10);			/* Completam les dades que falten a l'estructura */
	no.alfa=strtoul(row[1], NULL, 10);
	no.ghost=0;
	no.dirty=0;
	no.used=0;
	add_obert();								/* Afegim la nova entrada a la cach */
	rmo++;
	pthread_mutex_unlock(&mo);
	return 1;									/* I avisam de que hi es */
}




/****** Funcions que proporciona la llibreria - Accs a MySQL amb cach - Gravador, Stats i Servidor ******/

/* Aquest s el codi del thread "gravador" */
inline void gravador (void)
{
	register int			temps;
	register u_long			hash;
	register NodeHost		*tmph;
	register NodeObert		*tmpo;
	/* Resulta innecessari "register NodeServei *tmps;", no tocarem la taula perqu s "read only". */

	for(;;) {										/* Bucle "infinit" :-) */

		pthread_mutex_lock(&mf);					/* Abans de continuar, miram si el thread principal te res a dir */
		if (flag_acabar) {
			pthread_mutex_unlock(&mf);
			return;
		}
		pthread_mutex_unlock(&mf);

		temps=GRAVADOR_SLEEP;						/* Dormim una estona entre gravaci i gravaci */
		while(temps>0){
			temps=sleep(temps);
		}
	
#ifdef DEBUG
		printf("gravant: ");
#endif
		for(hash=0; hash<HASH_SIZE; hash++) {		/* Per cada possible hash, fem el mateix a totes les taules */

			/*** Taula HOST ***/
			pthread_mutex_lock(&mh);				/* Bloquejam la cache d'aquesta taula */
			tmph=hh[hash];							/* I comenant pel primer element amb aquest hash */
			while(tmph) {							/* Recorrem la llista enllaada de les colisions */
				if(tmph->ghost) {					/* Si es una nova entrada, fem un INSERT */
					db_insert_host(tmph);
					tmph->ghost=0;
					tmph->dirty=0;
				}
				if(tmph->dirty) {					/* Si no, i es una entrada modificada, fem un UPDATE */
					db_update_omega_host(tmph);
					tmph->dirty=0;
				}
				tmph=tmph->next;
			}
			pthread_mutex_unlock(&mh);				/* Desbloquejam la cache d'aquesta taula */

			/*** Taula OBERT ***/
			pthread_mutex_lock(&mo);				/* Bloquejam la cache d'aquesta taula */
			tmpo=ho[hash];							/* I comenant pel primer element amb aquest hash */
			while(tmpo) {							/* Recorrem la llista enllaada de les colisions */
				if(tmpo->ghost) {					/* Si es una nova entrada, fem un INSERT */
					db_insert_obert(tmpo);
					tmpo->ghost=0;
					tmpo->dirty=0;
				}
				if(tmpo->dirty) {					/* Si no, i es una entrada modificada, fem un UPDATE */
					db_update_omega_obert(tmpo);
					tmpo->dirty=0;
				}
				tmpo=tmpo->next;
			}
			pthread_mutex_unlock(&mo);				/* Desbloquejam la cache d'aquesta taula */
	
			/*** Taula SERVEI ***/
			/* Per sort per la meva salut mental ;-) aquesta taula es "read only". */
			/* Aqui no hem de fer res mes, continuam amb el proper hash. */
		}
#ifdef DEBUG
		printf("ok\n");
#endif
	}
}


/* Aquest s el codi del thread "stats" */
inline void stats (void)
{
	int temps;
	int rhh_prev = 0;
	int rmh_prev = 0;
	int rho_prev = 0;
	int rmo_prev = 0;
	int rh_diff;
	int ro_diff;

	for(;;) {										/* Bucle "infinit" :-) */

		pthread_mutex_lock(&mf);					/* Abans de continuar, miram si el thread principal te res a dir */
		if (flag_acabar) {
			pthread_mutex_unlock(&mf);
			return;
		}
		pthread_mutex_unlock(&mf);

		temps=STATS_SLEEP;							/* Dormim una estona entre stats i stats :-) */
		while(temps>0){
			temps=sleep(temps);
		}
	
#ifdef DEBUG
		printf("stats: ");
#endif
		pthread_mutex_lock(&mh);
		rh_ea=((rh_diff=(rhh-rhh_prev)+(rmh-rmh_prev))?(ALPHA*((rhh-rhh_prev)/(float)(rh_diff))):(0))+(1.0-ALPHA)*(rh_ea);
		rhh_prev = rhh;
		rmh_prev = rmh;
		pthread_mutex_unlock(&mh);

		pthread_mutex_lock(&mo);
		ro_ea=((ro_diff=(rho-rho_prev)+(rmo-rmo_prev))?(ALPHA*((rho-rho_prev)/(float)(ro_diff))):(0))+(1.0-ALPHA)*(ro_ea);
		rho_prev = rho;
		rmo_prev = rmo;
		pthread_mutex_unlock(&mo);
#ifdef DEBUG
		printf("ok\n");
#endif
	}
}


/* Aquest s el codi del thread "servidor" */
inline void servidor (void)
{
	u_char				pbuf[MAX_QUERY_LEN], qbuf[MAX_QUERY_LEN];
	struct sockaddr_in	sin;
	int					m = 0, s = 0, yes = 1;

	memset(&sin, 0, sizeof(sin));
	sin.sin_family=AF_INET;
	sin.sin_port=htons(PORT_SERVIDOR);
	sin.sin_addr.s_addr=htonl(INADDR_ANY);
	if((m=socket(PF_INET, SOCK_STREAM, 0))<0)
		PANIC("socket");
	if(setsockopt(m, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(yes))<0)
		PANIC("setsockopt");
	if(bind(m, (struct sockaddr *)&sin, sizeof(sin))<0)
		PANIC("bind");
	if(listen(m, QLEN)<0)
		PANIC("listen");

	for(;;) {										/* Bucle "infinit" :-) */

		pthread_mutex_lock(&mf);					/* Abans de continuar, miram si el thread principal te res a dir */
		if (flag_acabar) {
			pthread_mutex_unlock(&mf);
			close(m);
			return;
		}
		pthread_mutex_unlock(&mf);

		s=accept(m, NULL, NULL);
	
#ifdef DEBUG
		printf("servidor: ");
#endif
		read(s, qbuf, sizeof(qbuf));

		if (!strncmp(qbuf, GET_stats_xsl, strlen(GET_stats_xsl))) {				/* si demanen stats.xsl, el servim */
			snprintf(pbuf, sizeof(pbuf), xsl_stats);
		}
		else {																	/* si no, servim stats.xml */
			snprintf(pbuf, sizeof(pbuf), xml_stats_begin);

			pthread_mutex_lock(&mh);
			snprintf(qbuf, sizeof(qbuf), xml_stat_cache, "host", rhh, rmh, rh_ea, FULL(htlenh));
			pthread_mutex_unlock(&mh);
			strncat(pbuf, qbuf, sizeof(pbuf));

			pthread_mutex_lock(&mo);
			snprintf(qbuf, sizeof(qbuf), xml_stat_cache, "obert", rho, rmo, ro_ea, FULL(htleno));
			pthread_mutex_unlock(&mo);
			strncat(pbuf, qbuf, sizeof(pbuf));

			strncat(pbuf, xml_stats_end, sizeof(pbuf));
		}

		write(s, pbuf, strnlen(pbuf, sizeof(pbuf)));
		close(s);
#ifdef DEBUG
		printf("ok\n");
#endif
	}
}




