/*
 * dbcache.h, PFC d'en Guillem Cantallops Ramis a la UIB
 *
 * Copyright (C) 2000-2001, GCR.
 *
 * Distributed under the terms of the GNU GPL
 *
 */



/* Funci per mostrar alguns errors del MySQL*/

#define MYSQL_ERR(M,N)  { printf("%s[%d]: %s: %s\n", __FILE__, __LINE__, N, mysql_error(M)); exit(-1); }

/* Gesti d'errors a algunes system calls */
 
#define PANIC(SCALL)    { printf("%s[%d]: %s: %s\n", __FILE__, __LINE__, SCALL, strerror(errno)); exit(-1); }



/* Constants per la connexio a la BD MySQL, amb noms molt descriptius ;-D */
 
#define HOST		"localhost"
#define USER		"guillem"
#define PASSWD		"adfgvx"
#define DB			"monitor"
#define PORT		0
#define UNIX_SOCKET	"/var/lib/mysql/mysql.sock"
#define CLIENT_FLAG	0
 
#define MAX_QUERY_LEN	1024	/* Aix s una cosa meva, no ho limita el MySQL */



/* Temps (en segons) que passara dormit el thread gravador entre gravaci i gravaci */

#define GRAVADOR_SLEEP	17		/* Usant nombres primers, no solen coincidir ;-) */
#define STATS_SLEEP		23
#define SERVIDOR_SLEEP	31



/* Port a on escoltar el servidor */

#define PORT_SERVIDOR	1979
#define QLEN			5



/* Constants pel hashing */

#define HASH_BITS	8						/* Tamany del hash en bits (original: 10) */
#define HASH_SIZE	(1UL<<HASH_BITS)		/* La taula te 2^HASH_BITS entrades (original: 1024) */
#define HASH_MASK	(HASH_SIZE-1)			/* Mascara de HASH_BITS uns abaix (original: 0x3ff) */

#define MAX_LL_LEN	2						/* Maxim d'elements a un "bucket" (promig, si limitam el segent) */
#define MAX_HT_LEN	(HASH_SIZE*MAX_LL_LEN)	/* Maxim d'elements a la cach d'una taula (usat al CLOCK) */



/* HASH(KEY) simplement trunca la clau a un tamany adequat perque faci de hash,
   pero si fem servir varies claus ho podem arreglar molt facil i eficientment:
   noms li hem de passar la XOR de tots ells ;-) */

#define HASH(KEY)	(HASH_MASK&(KEY))

#define ALPHA		0.5



/* Aix s pel meu cas particular, per les taules que faig servir: */

#define HH(IP, MACH, MACL)						HASH(IP^MACH^MACL)
#define HO(IP, MACH, MACL, PORT, PROTOCOL)		HASH(IP^MACH^MACL^PORT^PROTOCOL)
#define HS(PORT, PROTOCOL)						HASH(PORT^PROTOCOL)



/* Aquesta s l'estructura per construir les taules de hash, igualment particularitzada.
   Usarem taules de HASH_SIZE elements d'aquests. Inicialitzar amb memset() a tot-zeros.*/

typedef struct node_host {
	u_long	ip;						/* Dades d'una entrada de la taula HOST */
	u_long	mach;
	u_long	macl;
	time_t	omega;
	time_t	alfa;
	u_char	ghost:1,				/* Dades noves -> encara s'han d'insertar a la BD */
			dirty:1,				/* Dades modificades -> s'han d'actualitzar a la BD */
			used:1,					/* Dades usades recentment -> per seleccionar victimes */
			reserved:5;				/* Aquests 5 bits sobren, de moment */
	struct node_host	*next;		/* Punter al segent element del bucket */
	struct node_host	*clnext;	/* Punter al segent element de la cach de la taula (usat al CLOCK) */
} NodeHost;

#define		NHKLEN				(3*sizeof(u_long))
#define		NHCMP(NH1, NH2)		(memcmp((const void *)NH1, (const void *)NH2, (size_t)NHKLEN))


typedef struct node_obert {
	u_long	ip;						/* Dades d'una entrada de la taula OBERT */
	u_long	mach;
	u_long	macl;
	u_short	port;
	u_char	protocol;
	time_t	omega;
	time_t	alfa;
	u_char	ghost:1,				/* Dades noves -> encara s'han d'insertar a la BD */
			dirty:1,				/* Dades modificades -> s'han d'actualitzar a la BD */
			used:1,					/* Dades usades recentment -> per seleccionar victimes */
			reserved:5;				/* Aquests 5 bits sobren, de moment */
	struct node_obert 	*next;		/* Punter al segent element del bucket */
	struct node_obert	*clnext;	/* Punter al segent element de la cach de la taula (usat al CLOCK) */
} NodeObert;

#define		NOKLEN				(3*sizeof(u_long)+sizeof(u_short)+sizeof(u_char))
#define		NOCMP(NO1, NO2)		(memcmp((const void *)NO1, (const void *)NO2, (size_t)NOKLEN))


typedef struct node_servei {
	u_short	port;					/* Dades d'una entrada de la taula SERVEI */
	u_char	protocol;
	struct node_servei	*next;		/* Punter al segent element del bucket */
} NodeServei;

#define		NSKLEN				(sizeof(u_short)+sizeof(u_char))
#define		NSCMP(NS1, NS2)		(memcmp((const void *)NS1, (const void *)NS2, (size_t)NSKLEN))

/* ALERTA: a node_servei ens hem deixat el camp "nom" de la taula "servei" i els flags perque no ens fan falta,
   la taula ser "read only", i a ms el camp aquest que no s'usa s de tipus "varchar" que s molt pesat :-P */



/* Macros per XML/XSL, servidor, estadistiques, etc. ;-D */

static const char *xml_stats_begin =
"HTTP/1.0 200 OK
Content-Type: text/xml
 
<?xml version='1.0'?>
<!DOCTYPE stats [
	<!ELEMENT stats (stat+)>
		<!ELEMENT stat (cache|captura)>
			<!ELEMENT cache (table, read_hits, read_misses, read_hit_ratio, full)>
				<!ELEMENT table (#PCDATA)>
				<!ELEMENT read_hits (#PCDATA)>
				<!ELEMENT read_misses (#PCDATA)>
				<!ELEMENT read_hit_ratio (#PCDATA)>
				<!ELEMENT full (#PCDATA)>
			<!ELEMENT captura (num_captures)>
				<!ELEMENT num_captures (#PCDATA)>
]>
<?xml:stylesheet href='/stats.xsl' type='text/xsl'?>
<stats>
";

static const char *xml_stat_cache =
"<stat>
	<cache>
		<table>%s</table>
		<read_hits>%u</read_hits>
		<read_misses>%u</read_misses>
		<read_hit_ratio>%4.3f</read_hit_ratio>
		<full>%3.1f%%</full>
	</cache>
</stat>
";

static const char *xml_stats_end =
"</stats>
";

static const char *xsl_stats =
"HTTP/1.0 200 OK
Content-Type: text/xsl
 
<?xml version='1.0'?>
<xsl:stylesheet xmlns:xsl='http://www.w3.org/TR/WD-xsl'>
	<xsl:template match='/'>
		<html>
			<head>
				<title>monitord stats</title>
			</head>
			<body>
				<h1>monitord stats</h1>
				<xsl:apply-templates select='//stat/*'/>
			</body>
		</html>
	</xsl:template>
	<xsl:template match='//stat/cache'>
		<table border='1'>
			<tr>
				<th>Cache</th>
			</tr>
			<tr>
				<td>Table</td>
				<td><xsl:value-of select='table'/></td>
			</tr>
			<tr>
				<td>Read hits</td>
				<td><xsl:value-of select='read_hits'/></td>
			</tr>
			<tr>
				<td>Read misses</td>
				<td><xsl:value-of select='read_misses'/></td>
			</tr>
			<tr>
				<td>Read hit ratio</td>
				<td><xsl:value-of select='read_hit_ratio'/></td>
			</tr>
			<tr>
				<td>Full</td>
				<td><xsl:value-of select='full'/></td>
			</tr>
		</table>
		<br/>
	</xsl:template>
</xsl:stylesheet>
";

#define		FULL(HT_LEN)	(100.0*(float)HT_LEN/(float)MAX_HT_LEN)

#define		GET_stats_xsl	"GET /stats.xsl"



/* Funcions que proporcionam (veure dbcache.c per ms detalls :-) */

inline void init_dbcache (void);
inline void close_dbcache (void);
inline void insertar_host (u_int ip, u_int mach, u_int macl);
inline void actualitzar_omega_host (u_int ip, u_int mach, u_int macl);
inline void insertar_obert (u_int ip, u_int mach, u_int macl, u_short port, u_char protocol);
inline void actualitzar_omega_obert (u_int ip, u_int mach, u_int macl, u_short port, u_char protocol);
inline int hi_es_el_host (u_int ip, u_int mach, u_int macl);
inline int hi_es_el_servei (u_short port, u_char protocol);
inline int esta_obert_i_ho_sabem (u_int ip, u_int mach, u_int macl, u_short port, u_char protocol);
inline void gravador (void);
inline void stats (void);
inline void servidor (void);


