/*
 * monitord.c, PFC d'en Guillem Cantallops Ramis a la UIB
 *
 * Copyright (C) 2000-2001, GCR.
 *
 * Distributed under the terms of the GNU GPL
 *
 */



#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <net/if.h>
#include <net/bpf.h>
#include <netpacket/packet.h>
#include <signal.h>
#include <pthread.h>
#include "monitord.h"
#include "dbcache.h"



/* Variables globals */

extern int					errno;

static int					raw;								/* descriptor pel raw socket */
static u_char				buf[MAX_FRAME_LEN];					/* buffer pel paquet capturat */

static u_int				rebuts=0;							/* comptador de paquets capturats */
static struct timeval		tx, ty;								/* temps inicial i temps final de la captura */

static pthread_t			th_capturador;						/* threads apart del principal */
static pthread_t			th_gravador;
static pthread_t			th_stats;
static pthread_t			th_servidor;
static sigset_t				set;								/* aix es per fer el sigwait() */
static int					sig;

int							flag_acabar = 0;					/* flag global per controlar l'acabament, amb mutex */
pthread_mutex_t				mf = PTHREAD_MUTEX_INITIALIZER;



/* Funci que s'executa sempre al principi: inicialitza parametres del capturador */

inline void init_capturador (u_char *device)
{
	struct ifreq		ifr;
	struct sockaddr_ll	sll;
	struct packet_mreq	pmr;
	struct bpf_insn		programa[] = {	/* no pens comentar aix, s entre el kernel i jo :-P */
		{40,	0,	0,	12		},
		{21,	12,	0,	2054	},
		{21,	11,	0,	32821	},
		{21,	0,	2,	2048	},
		{48,	0,	0,	23		},
		{21,	3,	9,	6		},
		{21,	0,	8,	34525	},
		{48,	0,	0,	20		},
		{21,	0,	6,	6		},
		{40,	0,	0,	20		},
		{69,	4,	0,	8191	},
		{177,	0,	0,	14		},
		{80,	0,	0,	27		},
		{21,	0,	1,	18		},
		{6,		0,	0,	65535	},
		{6,		0,	0,	0		}
	};
	struct bpf_program	filtre = {		/* aquest filtre deixa passar arp, rarp, i tcp syn+ack */
		sizeof(programa)/sizeof(struct bpf_insn),
		programa
	};

	/* Obtenim el descriptor de raw socket */
#ifdef DEBUG
	printf("socket: ");
#endif
	if((raw=socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL)))<0)
		PANIC("socket");
#ifdef DEBUG
	printf("ok\n");
#endif

	/* Obtenim l'index de l'interface que ens interessa */
	memset((void *)&ifr, 0, sizeof(ifr));
	strncpy(ifr.ifr_name, device, sizeof(ifr.ifr_name));
	ifr.ifr_name[sizeof(ifr.ifr_name)-1]=0;
#ifdef DEBUG
	printf("ioctl: ");
#endif
	if(ioctl(raw, SIOCGIFINDEX, &ifr)<0)
		PANIC("ioctl");
#ifdef DEBUG
	printf("ok\n");
#endif

	/* Lligam el socket a l'interface en questi */
	memset((void *)&sll, 0, sizeof(sll));
	sll.sll_family=AF_PACKET;
	sll.sll_ifindex=ifr.ifr_ifindex;
	sll.sll_protocol=htons(ETH_P_ALL);
#ifdef DEBUG
	printf("bind: ");
#endif
	if(bind(raw, (struct sockaddr *)&sll, sizeof(sll))<0)
		PANIC("bind");
#ifdef DEBUG
	printf("ok\n");
#endif

	/* El posam en mode promiscu */
	memset((void *)&pmr, 0, sizeof(pmr));
	pmr.mr_ifindex=ifr.ifr_ifindex;
	pmr.mr_type=PACKET_MR_PROMISC;
#ifdef DEBUG
	printf("setsockopt: ");
#endif
	if(setsockopt(raw, SOL_PACKET, PACKET_ADD_MEMBERSHIP, (void *)&pmr, (socklen_t)sizeof(pmr))<0)
		PANIC("setsockopt");
#ifdef DEBUG
	printf("ok\n");
#endif

	/* Li posam un LSF (Linux Socket Filter) definit en BPF (Berkeley Packet Filter) */
#ifdef DEBUG
	printf("setsockopt: ");
#endif
	if(setsockopt(raw, SOL_SOCKET, SO_ATTACH_FILTER, &filtre, sizeof(filtre))<0)
		PANIC("setsockopt");
#ifdef DEBUG
	printf("ok\n");
#endif
}



/* Funci que s'executa sempre al final: neteja i imprimeix estadistiques... */

inline void acabar (void)
{
	/* Activam el flag que diu als threads que es hora d'acabar */
	pthread_mutex_lock(&mf);
	flag_acabar=1;
	pthread_mutex_unlock(&mf);

	/* Esperam l'acabament del thread capturador */
	printf("pthread_join: thread capturador acabant... ");
	pthread_join(th_capturador, NULL);
	printf("ok\n");

	/* Esperam l'acabament del thread gravador */
	printf("pthread_join: thread gravador acabant... ");
	pthread_join(th_gravador, NULL);
	printf("ok\n");

	/* Esperam l'acabament del thread stats */
	printf("pthread_join: thread stats acabant... ");
	pthread_join(th_stats, NULL);
	printf("ok\n");

	/* Esperam l'acabament del thread servidor */
	printf("pthread_join: thread servidor acabant... ");
	pthread_join(th_servidor, NULL);
	printf("ok\n");

	/* Tancam el raw socket */
#ifdef DEBUG
	printf("close: ");
#endif
	close(raw);
#ifdef DEBUG
	printf("ok\n");
#endif

	/* Tancam BD i cache (gestiona errors) */
	close_dbcache();

	/* I mostram els resultats */
	printf("En promig s'ha capturat un paquet cada %f segons\n",
		(TV(ty)-TV(tx))/((float)rebuts));

	/* That's all, folks! (aqui acaba realment el programa; es un lloc original, eh? ;-) */
	exit(0);
}



/* Funci cridada per processar_captura() cada vegada que reb un missatge ARP */

inline void processar_arp (const u_char *sp)
{
	arp_message *m;						/* 'm' de Missatge */
	u_int		ip, mach, macl;			/* Les adreces en format binari */
#ifdef DEBUG
	char		s_ip[IP_TXT_LEN];		/* Les adreces en format "string" */
	char		s_ha[HA_TXT_LEN];
#endif

	/* Extraiem les adreces dels camps corresponents */
	m=(arp_message *)(sp+ETH_HDR_LEN);
	ip=B_S_IP(m);
	mach=B_S_MACH(m);
	macl=B_S_MACL(m);
#ifdef DEBUG
	sprintf(s_ip,IP_FORMAT,S_IP(m));		/* Ara 's_ip' s una "string" amb l'adrea IP */
	sprintf(s_ha,HA_FORMAT,S_HA(m));		/* Ara 's_ha' s una "string" amb l'adrea Ethernet */
#endif

	/* Consultam i actualitzam la BD */
	if(!hi_es_el_host(ip, mach, macl)) {
		insertar_host(ip, mach, macl);				/* No hi es: insertar nova entrada */
#ifdef DEBUG
		printf("ARP %15s = %17s\tNOU!!!\n", s_ip, s_ha);
#endif
	}
	else {
		actualitzar_omega_host(ip, mach, macl);		/* Hi es: actualitzar timestamp del darrer avistament */
#ifdef DEBUG
		printf("ARP %15s = %17s\n", s_ip, s_ha);
#endif
	}
}



/* Funci cridada per processar_captura() cada vegada que reb un segment TCP */

inline void processar_tcp (const u_char *sp)
{
	eth_hdr		*t;					/* "t" de Trama */
	ip_hdr		*p;					/* "p" de Paquet */
	tcp_hdr		*s;					/* "s" de Segment */
	u_int		mach, macl, ip;		/* Adreces MAC i IP _del_servidor_ (origen del SYN+ACK) */
	u_short		port;				/* Port d'origen del SYN+ACK */
	u_char		protocol;			/* Protocol de transport. Aqui sempre hauria d'esser IP_P_TCP=6 */
	
#ifdef DEBUG
	char		s_ip[IP_TXT_LEN];		/* Les adreces IP d'origen i dest, en format "string" */
	char		t_ip[IP_TXT_LEN];
#endif

	/* Amb aix podrem accedir facilment als camps de les capaleres */
	t=(eth_hdr *)(sp);
	p=(ip_hdr *)(sp+ETH_HDR_LEN);
	s=(tcp_hdr *)(sp+ETH_HDR_LEN+IP_HDR_LEN);

#ifdef DEBUG
	sprintf(s_ip,IP_FORMAT,S_IP(p));
	sprintf(t_ip,IP_FORMAT,T_IP(p));
//	printf("TCP %15s:%-5u -> %15s:%-5u [%c%c%c%c%c%c]\t",
//		s_ip, ntohs(s->s_port), t_ip, ntohs(s->t_port),
//		(s->urg)?('U'):('u'), (s->ack)?('A'):('a'), (s->psh)?('P'):('p'),
//		(s->rst)?('R'):('r'), (s->syn)?('S'):('s'), (s->fin)?('F'):('f') );
#endif

	if (TCP_FLAGS(s)==(TCP_SYN|TCP_ACK)) {	/* De moment noms capturam SYN+ACK. Si arriba aix... */
		mach=B_S_MACH(t);					/* miram _qui_ accepta una connexi (i.e. _qui_ envia un SYN+ACK) */
		macl=B_S_MACL(t);
		ip=B_S_IP(p);
		port=ntohs(s->s_port);
		protocol=p->proto;
		if (hi_es_el_host(ip, mach, macl)) {			/* Si el tenim a la BD, es de la xarxa local i ens interessa! */
			if (hi_es_el_servei(port, protocol)) {		/* Pero nomes certs serveis */
				if (esta_obert_i_ho_sabem(ip, mach, macl, port, protocol)) {	/* Actualitzam TimeStamps */
					actualitzar_omega_obert(ip, mach, macl, port, protocol);
#ifdef DEBUG
					printf("TCP %15s:%-5u\tObert\n", s_ip, port);
#endif
				}
				else {															/* O afegim nova entrada */
					insertar_obert(ip, mach, macl, port, protocol);
#ifdef DEBUG
					printf("TCP %15s:%-5u\tObert I NOU!!!\n", s_ip, port);
#endif
				}
			}
//#ifdef DEBUG
//			else {
//				printf("Falsa alarma, servei desconegut que no interessa. Ignorant rpidament.\n");
//			}
//#endif
		}
//#ifdef DEBUG
//		else {											/* Si no, probablement s de fora i passam rpidament d'ell */
//			printf("Falsa alarma, host extern que no interessa. Ignorant rpidament.\n");
//		}
//#endif
	}
}



/* Funci cridada per processar_captura() cada vegada que reb un datagrama UDP

inline void processar_udp (const u_char *sp)
{
	ip_hdr *p;
	udp_hdr *d;

	p=(ip_hdr *)(sp+ETH_HDR_LEN);
	d=(udp_hdr *)(sp+ETH_HDR_LEN+IP_HDR_LEN);

#ifdef DEBUG
	printf("UDP ");
	printf(IP_FORMAT, S_IP(p));
	printf(":%u\t", ntohs(d->s_port));
	printf(IP_FORMAT, T_IP(p));
	printf(":%u\n", ntohs(d->t_port));
#endif
}
*/



/* Funci cridada cada vegada que pillam un paquet */

inline void processar_captura (const u_char *sp)
{
	eth_hdr *t;		/* "t" de Trama */
	ip_hdr *p;		/* "p" de Paquet */

	/* Si es la primera captura, guardam el seu TimeStamp: comenam a comptar */
	if(!rebuts)
		gettimeofday(&tx, NULL);

	/* Incrementam comptador de paquets rebuts */
	rebuts++;

	/* Guardam el TimeStamp del darrer paquet capturat */
	gettimeofday(&ty, NULL);

	/* Anem a veure qu diu la capalera Ethernet sobre l'info que du la trama */
	t=(eth_hdr *)sp;
	switch (ntohs(t->type)) {
		case ETH_P_ARP:						/* Tant si s ARP */
		case ETH_P_RARP:					/* com si s RARP */
			processar_arp(sp);				/* els processam igual */
			break;
		case ETH_P_IP:						/* Si s IP, encara hem d'afinar ms */
			p=(ip_hdr *)(sp+ETH_HDR_LEN);	/* Anem a veure que diu la capalera IP sobre l'info que du el paquet */
			switch (p->proto) {
				case IP_P_TCP:				/* Si s TCP */
					processar_tcp(sp);		/* ja sabem el que hem de fer */
					break;
//				case IP_P_UDP:
//					processar_udp(sp);
//					break;
				default:
					break;
			}
			break;
		default:
			break;
	}
}



/* Aquest es el codi del thread "capturador" */
inline void capturador (void)
{
	int rebut;

	for(;;) {
		do {
			rebut=recvfrom(raw, buf, sizeof(buf), MSG_TRUNC, NULL, 0);		/* Capturam un paquet */
		} while ((rebut==-1)&&(errno==EINTR));								/* Evitant fals desbloqueig */

		if(rebut==-1) {														/* Els altres errors son fatals */
			PANIC("recvfrom");
		}

		processar_captura(buf);												/* Si arribam aqui, el processam */

		pthread_mutex_lock(&mf);							/* Miram si el thread principal ens ha dit que acabem */
		if (flag_acabar) {
			pthread_mutex_unlock(&mf);
			return;
		}
		pthread_mutex_unlock(&mf);
	}
}



/* Programa Principal! */

int main (int argc, char **argv)
{
	/* IMPORTANT: fem stdout "unbuffered". NO posar cap instrucci abans d'aquesta. */
	setvbuf(stdout, (char *)NULL, _IONBF, 0);

	/* Inicialitzam capturador, i desprs deixam els privilegis de root que ja no fan falta */
	init_capturador(((argc==2)&&(strlen(argv[1])>0))?(argv[1]):(DEFAULT_DEVICE));
	setuid(getuid());
	
	/* Inicialitzam BD i cache (gestiona errors) */
	init_dbcache();

	/* NOTA: A partir de l'implementaci amb LinuxThreads, no farem servir senyals assincrones, */
	/* mes que res, perqu ens pot quedar el thread penjat :-P Ara fem servir sigwait(), veure abaix. */

	/* Acabarem "civilitzadament" si rebem aquests senyals */
	sigaddset(&set, SIGINT);
	sigaddset(&set, SIGQUIT);
	sigaddset(&set, SIGTERM);
	pthread_sigmask(SIG_BLOCK, &set, (sigset_t *)NULL);

	/* Ja podem comenar a capturar paquets */
	pthread_create(&th_capturador, NULL, (void *)capturador, (void *)NULL);

	/* I a gravar peridicament les modificacions que fem a la cach */
	pthread_create(&th_gravador, NULL, (void *)gravador, (void *)NULL);

	/* Tamb comenam a actualitzar les estadistiques */
	pthread_create(&th_stats, NULL, (void *)stats, (void *)NULL);

	/* I a servir-les */
	pthread_create(&th_servidor, NULL, (void *)servidor, (void *)NULL);

	/* Esperam que ens aturin */
	sigwait(&set, &sig);

	/* Ooops! Hem rebut SIGINT, SIGQUIT  SIGTERM: acabam de forma "civilitzada" :-) */
	acabar();

	/* Mai arribam aqu, per aix evitam el "warning: control reaches end of non-void function" */
	exit(-1);
}


